@@ -4,6 +4,7 @@
  * This sets/verifies the capabilities of a given file.
  */
 
+#include <ctype.h>
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
@@ -24,15 +25,45 @@ static void usage(int status)
 	    " [ Note: capsh --suggest=\"something...\" might help you pick. ]"
 	    "\n"
 	    " -h          this message and exit status 0\n"
+	    " -f          force setting even when the capability is invalid\n"
 	    " -q          quietly\n"
 	    " -v          validate supplied capability matches file\n"
-	    " -n <rootid> write a user namespace limited capability\n"
+	    " -n <rootid> write a user namespace (!= 0) limited capability\n"
 	    " --license   display the license info\n"
 	);
     exit(status);
 }
 
-#define MAXCAP  2048
+/* parse a positive integer with some error handling */
+static unsigned long pos_uint(const char *text, const char *prefix, int *ok)
+{
+    char *remains;
+    unsigned long value;
+    ssize_t len = strlen(text);
+
+    if (len == 0 || *text == '-') {
+	goto fail;
+    }
+    value = strtoul(text, &remains, 0);
+    if (*remains || value == 0) {
+	goto fail;
+    }
+    if (ok != NULL) {
+	*ok = 1;
+    }
+    return value;
+
+fail:
+    if (ok == NULL) {
+	fprintf(stderr, "%s: want positive integer, got \"%s\"\n",
+		prefix, text);
+	exit(1);
+    }
+    *ok = 0;
+    return 0;
+}
+
+#define MAXCAP 2048
 
 static int read_caps(int quiet, const char *filename, char *buffer)
 {
@@ -70,7 +101,7 @@ int main(int argc, char **argv)
 {
     int tried_to_cap_setfcap = 0;
     char buffer[MAXCAP+1];
-    int retval, quiet = 0, verify = 0;
+    int retval, quiet = 0, verify = 0, forced = 0;
     cap_t mycaps;
     cap_value_t capflag;
     uid_t rootid = 0, f_rootid;
@@ -85,9 +116,12 @@ int main(int argc, char **argv)
 		" (old libcap?)\n");
     }
 
+    cap_t cap_d = NULL;
     while (--argc > 0) {
 	const char *text;
-	cap_t cap_d;
+
+	cap_free(cap_d);
+	cap_d = NULL;
 
 	if (!strcmp(*++argv, "-q")) {
 	    quiet = 1;
@@ -96,10 +130,14 @@ int main(int argc, char **argv)
 	if (!strcmp("--license", *argv)) {
 	    printf(
 		"%s see LICENSE file for details.\n"
-		"Copyright (c) 1997,2007-8,2020,21 Andrew G. Morgan"
+		"Copyright (c) 1997,2007-8,2020-21 Andrew G. Morgan"
 		" <morgan@kernel.org>\n", argv[0]);
 	    exit(0);
 	}
+	if (!strcmp(*argv, "-f")) {
+	    forced = 1;
+	    continue;
+	}
 	if (!strcmp(*argv, "-h")) {
 	    usage(0);
 	}
@@ -109,19 +147,17 @@ int main(int argc, char **argv)
 	}
 	if (!strcmp(*argv, "-n")) {
 	    if (argc < 2) {
-		fprintf(stderr, "usage: .. -n <rootid> .. - rootid!=0 file caps");
+		fprintf(stderr,
+			"usage: .. -n <rootid> .. - rootid!=0 file caps");
 		exit(1);
 	    }
 	    --argc;
-	    rootid = (uid_t) atoi(*++argv);
-	    if (rootid+1 < 2) {
-		fprintf(stderr, "invalid rootid!=0 of '%s'", *argv);
-		exit(1);
-	    }
+	    rootid = (uid_t) pos_uint(*++argv, "bad ns rootid", NULL);
 	    continue;
 	}
 
 	if (!strcmp(*argv, "-r")) {
+	    cap_free(cap_d);
 	    cap_d = NULL;
 	} else {
 	    if (!strcmp(*argv,"-")) {
@@ -133,6 +169,17 @@ int main(int argc, char **argv)
 		text = *argv;
 	    }
 
+	    int non_space = 0, j;
+	    for (j = 0; text[j]; j++) {
+		if (!isspace(text[j])) {
+		    non_space = 1;
+		    break;
+		}
+	    }
+	    if (!non_space) {
+		fprintf(stderr, "empty space is an invalid capability, did you mean -r?\n");
+		exit(1);
+	    }
 	    cap_d = cap_from_text(text);
 	    if (cap_d == NULL) {
 		perror("fatal error");
@@ -144,11 +191,9 @@ int main(int argc, char **argv)
 	    }
 #ifdef DEBUG
 	    {
-		ssize_t length;
-		const char *result;
-
-		result = cap_to_text(cap_d, &length);
+		char *result = cap_to_text(cap_d, NULL);
 		fprintf(stderr, "caps set to: [%s]\n", result);
+		cap_free(result);
 	    }
 #endif
 	}
@@ -163,13 +208,20 @@ int main(int argc, char **argv)
 	    int cmp;
 
 	    if (cap_d == NULL) {
-		cap_d = cap_from_text("=");
+		cap_d = cap_init();
+		if (cap_d == NULL) {
+		    perror("unable to obtain empty capability");
+		    exit(1);
+		}
 	    }
 
 	    cap_on_file = cap_get_file(*++argv);
-
 	    if (cap_on_file == NULL) {
-		cap_on_file = cap_from_text("=");
+		cap_on_file = cap_init();
+		if (cap_on_file == NULL) {
+		    perror("unable to use missing capability");
+		    exit(1);
+		}
 	    }
 
 	    cmp = cap_compare(cap_on_file, cap_d);
@@ -210,12 +262,11 @@ int main(int argc, char **argv)
 		}
 		tried_to_cap_setfcap = 1;
 	    }
-	    retval = cap_set_file(*++argv, cap_d);
-	    if (retval != 0) {
-		int explained = 0;
-		int oerrno = errno;
-		int somebits = 0;
 #ifdef linux
+	    {
+		/* Linux's file capabilities have a compressed representation. */
+		int explained = 0;
+		int somebits = 0;
 		cap_value_t cap;
 		cap_flag_value_t per_state;
 
@@ -234,12 +285,18 @@ int main(int argc, char **argv)
 		    }
 		}
 		if (somebits && explained) {
-		    fprintf(stderr, "NOTE: Under Linux, effective file capabilities must either be empty, or\n"
-			    "      exactly match the union of selected permitted and inheritable bits.\n");
+		    fprintf(stderr, "Error: under Linux, effective file capabilities must either be empty, or\n"
+			    "       exactly match the union of selected permitted and inheritable bits.\n");
+		    if (!forced) {
+			exit(1);
+		    }
 		}
+	    }
 #endif /* def linux */
-
-		switch (oerrno) {
+	    errno = 0;
+	    retval = cap_set_file(*++argv, cap_d);
+	    if (retval != 0) {
+		switch (errno) {
 		case EINVAL:
 		    fprintf(stderr,
 			    "Invalid file '%s' for capability operation\n",
@@ -250,19 +307,23 @@ int main(int argc, char **argv)
 			fprintf(stderr,
 				"File '%s' has no capablity to remove\n",
 				argv[0]);
+			if (forced) {
+			    break;
+			}
 			exit(1);
 		    }
+		    /* FALLTHROUGH */
 		default:
 		    fprintf(stderr,
 			    "Failed to set capabilities on file '%s': %s\n",
-			    argv[0], strerror(oerrno));
+			    argv[0], strerror(errno));
 		    exit(1);
 		}
 	    }
 	}
-	if (cap_d) {
-	    cap_free(cap_d);
-	}
+    }
+    if (cap_d) {
+	cap_free(cap_d);
     }
 
     exit(0);
